import pkg_resources

from .exhaustive_search import ExhaustiveLinearRegression
from .integrate import (
    integrate_log_values_in_line,
    integrate_log_values_in_square,
    validate_list_dimension,
)
from .linear_regression import (
    ConstantRegression,
    LinearRegression,
    MarginalLinearRegression,
    MarginalNullRegression,
)
from .plot import feature_posterior, sigma_posterior, weight_diagram
from .probabilities import RandomVariable, gamma, inverse, uniform
from .scaler import StandardScaler

try:
    __version__ = pkg_resources.get_distribution("exhbma").version
except pkg_resources.DistributionNotFound:
    # Only for document generation
    __version__ = "dev"



import numpy as np


class StandardScaler(object):
    def __init__(self, n_dim, scaling: bool = True):
        self.n_dim = n_dim
        self.scaling = scaling

    def fit(self, array):
        if self.n_dim == 1:
            self.mean = np.mean(array)
            self.std = np.mean((array - self.mean) ** 2) ** 0.5
        elif self.n_dim == 2:
            self.mean = np.mean(array, axis=0)
            self.std = np.mean((array - self.mean) ** 2, axis=0) ** 0.5

    def transform(self, array):
        ret = np.array(array) - self.mean
        if self.scaling:
            ret = ret / self.std
        return ret

    def restore(self, array):
        ret = np.array(array)
        if self.scaling:
            ret = ret * self.std
        ret = ret + self.mean
        return ret



from typing import List, Optional

import numpy as np
from pydantic import BaseModel, Field
from scipy.stats import gamma as sp_gamma


class RandomVariable(BaseModel):
    position: float = Field(..., description="data position")
    prob: float = Field(..., ge=0, description="probability density/mass at position")


def gamma(
    x: np.ndarray,
    low: Optional[float] = None,
    high: Optional[float] = None,
    shape: float = 1e-3,
    scale: float = 1e3,
) -> List[RandomVariable]:
    """
    Gamma distribution: x ~ (const.) * x^(shape - 1) * exp(- x / scale)
    Distribution is limited on range of [low, high].
    If low(high) is None, low(high) is set as the minimum(maximum) value of x.
    """
    if low is None:
        low = min(x)
    if high is None:
        high = max(x)
    rv = sp_gamma(a=shape, scale=scale)
    norm = rv.cdf(x=high) - rv.cdf(x=low)
    probs = rv.pdf(x=x) / norm
    return [RandomVariable(position=i, prob=p) for (i, p) in zip(x, probs)]


def uniform(x: np.ndarray, low: float = 0.0, high: float = 1.0) -> List[RandomVariable]:
    """
    Uniform distribution: p(x) = 1 / (high - low)
    """
    prob = 1 / (high - low)
    return [RandomVariable(position=position, prob=prob) for position in x]


def inverse(
    x: np.ndarray,
    low: Optional[float] = None,
    high: Optional[float] = None,
) -> List[RandomVariable]:
    if low is None:
        low = min(x)
    if high is None:
        high = max(x)

    norm = np.log(high) - np.log(low)
    probs = (1 / x) / norm
    return [RandomVariable(position=i, prob=p) for (i, p) in zip(x, probs)]



import logging
from enum import Enum, auto
from itertools import product
from typing import List, Tuple, Union

import numpy as np
from pydantic import BaseModel, Field
from scipy.special import logsumexp
from tqdm import tqdm

from exhbma.linear_regression import (
    LinearRegression,
    MarginalLinearRegression,
    MarginalNullRegression,
)
from exhbma.probabilities import RandomVariable

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class ModelInfo(BaseModel):
    indicator: List[int] = Field(
        ...,
        description="Indicator vector of the model. This attribute may be excluded in the future, please use parent's `indicators_` instead.",  # noqa
    )
    log_prior: float = Field(
        ...,
        description="Log-prior of the model. This attribute may be excluded in the future, please use parent's `log_priors_` instead.",  # noqa
    )
    coefficient: List[float] = Field(
        ...,
        description="Coefficient of linear model, which is marginalized over sigma_noise and sigma_coef.",  # noqa
    )
    log_likelihood: float = Field(
        ..., description="Marginal log-likelihood of the model."
    )
    log_likelihood_over_sigma: List[List[float]] = Field(
        ...,
        description="Log-likelihood over sigma_noise and sigma_coef, `p(y| sigma_noise, sigma_coef, X)`.",  # noqa
    )


class PredictMode(Enum):
    select = auto()
    full = auto()


class ExhaustiveLinearRegression(object):
    r"""
    ExhaustiveSearchModel with linear_model.LinearRegression

    - Intercept of the linear model is assumed to be zero.

        - Assume that target variable y is centralized.
        - Assume that all features x are centralized and normalized.

    Parameters
    ----------
    sigma_noise_points: List[RandomVariable]
        Data points to explore sigma_noise parameter in exhaustive search.

    sigma_coef_points: List[RandomVariable]
        Data points to explore sigma_coef parameter in exhaustive search.

    alpha: float (default: 0.5)
        Alpha parameter is fixed to this value.

    exclude_null: bool (default: False)
        Whether or not exclude a null model.

    Attributes
    ----------
    n_features_in_: int
        Number of features seen during fit.

    coef_: List[float]
        Coefficients of the regression model (mean of distribution).

    log_likelihood_: float
        Log-likelihood of the model.
        Marginalization is performed over sigma_noise, sigma_coef, indicators.

    log_likelihood_over_sigma_: List[List[float]]
        Log-likelihood over :math:`\sigma_{noise}` and :math:`\sigma_{coef}`,
        :math:`p(y| \sigma_{noise}, \sigma_{coef}, X)`,
        which is marginalized over indicators.
        Prior distributions for both sigma are not included.

    feature_posteriors_: List[float]
        Posterior probabilities for each feature.

    indicators_: List[List[int]]
        List of indicator vectors.
        Null model `[0, 0, ..., 0]` are excluded,
        so length is :math:`2^{n\_features\_in\_} - 1`.

    log_priors_: List[float]
        List of log-prior probabilities for each model specified by indicator.

    log_likelihoods_: List[float]
        List of log-likelihood of each model specified by indicator.

    models_: List[ModelInfo]
        Information for all models specified by indicator vector.
        Length of this attribute is equal to that of indicators_
        and models correspond to each other.
    """

    def __init__(
        self,
        sigma_noise_points: List[RandomVariable],
        sigma_coef_points: List[RandomVariable],
        alpha: float = 0.5,
        exclude_null: bool = False,
    ):
        self.sigma_noise_points = sigma_noise_points
        self.sigma_coef_points = sigma_coef_points
        self.alpha = alpha
        self.exclude_null = exclude_null
        self._preprocessing_tolerance = 1e-8

    def fit(self, X: np.ndarray, y: np.ndarray, verbose: bool = True):
        """Train a model

        1. Create indicator vectors
        2. Fit sub-models for each indicator vector
           with marginalizing sigma_noise and sigma_coef for each indicator
        3. Calculate final model averaged over sub-models

        Parameters
        ----------
        X : np.ndarray with shape (n_data, n_features)
            Feature matrix. Each row corresponds to single data.

        y : np.ndarray with shape  (n_data,)
            Target value vector.

        verbose: bool (default: True)
            If this is set to `True`, progress bar is displayed.
        """
        self.n_features_in_: int = X.shape[1]

        LinearRegression.validate_target_centralization(
            y=y, tolerance=self._preprocessing_tolerance
        )
        LinearRegression.validate_feature_standardization(
            X=X, tolerance=self._preprocessing_tolerance
        )

        self.indicators_: List[List[int]] = self._generate_indicator(
            n_features=self.n_features_in_
        )
        self.log_priors_: List[float] = []
        self.log_likelihoods_: List[float] = []
        self.models_: List[ModelInfo] = []
        for indicator in tqdm(self.indicators_, disable=not verbose):
            (
                log_likelihood,
                log_likelihood_over_sigma,
                coefficient,
            ) = self._fit_over_sigma_noise_and_coef(
                X=X[:, np.array(indicator) == 1], y=y
            )
            self.log_priors_.append(self._fixed_alpha_prior(indicator=indicator))
            self.log_likelihoods_.append(log_likelihood)
            self.models_.append(
                ModelInfo(
                    indicator=indicator,
                    log_prior=self.log_priors_[-1],
                    coefficient=coefficient,
                    log_likelihood_over_sigma=log_likelihood_over_sigma,
                    log_likelihood=log_likelihood,
                )
            )

        self.log_likelihood_: float = self._calculate_log_marginal_likelihood(
            log_priors=self.log_priors_, models=self.models_
        )
        self.feature_posteriors_: List[float] = self._calculate_feature_posterior(
            log_priors=self.log_priors_,
            indicators=self.indicators_,
            models=self.models_,
        )
        self.log_likelihood_over_sigma_: List[
            List[float]
        ] = self._calculate_log_marginal_likelihood_over_sigma(
            log_priors=self.log_priors_, models=self.models_
        )

        coefficient = self._calculate_marginal_linear_model(
            log_priors=self.log_priors_,
            indicators=self.indicators_,
            models=self.models_,
        )
        self.coef_: List[float] = coefficient

    def _fit_over_sigma_noise_and_coef(
        self, X, y
    ) -> Tuple[float, List[List[float]], List[float]]:
        """
        Fit over (sigma_noise, sigma_coef) grid points and
        calculate log model likelihood by marginalizing.
        """
        model: Union[MarginalNullRegression, MarginalLinearRegression] = (
            MarginalLinearRegression(
                sigma_noise_points=self.sigma_noise_points,
                sigma_coef_points=self.sigma_coef_points,
            )
            if len(X[0]) > 0
            else MarginalNullRegression(
                sigma_noise_points=self.sigma_noise_points,
                sigma_coef_points=self.sigma_coef_points,
            )
        )
        model.fit(X=X, y=y, skip_preprocessing_validation=True)

        return (
            model.log_likelihood_,
            model.log_likelihood_over_sigma_,
            model.coef_,
        )

    def _calculate_log_marginal_likelihood(
        self, log_priors: List[float], models: List[ModelInfo]
    ) -> float:
        log_likelihood = logsumexp(
            [p + m.log_likelihood for (p, m) in zip(log_priors, models)]
        )
        return log_likelihood

    def _calculate_feature_posterior(
        self,
        log_priors: List[float],
        indicators: List[List[int]],
        models: List[ModelInfo],
    ) -> List[float]:
        log_joint_probabilities = [
            p + m.log_likelihood for (p, m) in zip(log_priors, models)
        ]
        log_marginal_likelihood = self._calculate_log_marginal_likelihood(
            log_priors=log_priors, models=models
        )
        np_indicators = np.array(indicators)
        log_marginals = []
        for column in range(np_indicators.shape[1]):
            indicator_values = np_indicators[:, column]
            log_marginals.append(
                logsumexp(a=log_joint_probabilities, b=indicator_values)
                - log_marginal_likelihood
            )
        return np.exp(log_marginals).tolist()

    def _calculate_log_marginal_likelihood_over_sigma(
        self, log_priors: List[float], models: List[ModelInfo]
    ) -> List[List[float]]:
        log_likelihood_over_sigma = logsumexp(
            [
                np.array(m.log_likelihood_over_sigma) + p
                for (m, p) in zip(models, log_priors)
            ],
            axis=0,
        )
        return log_likelihood_over_sigma.tolist()

    def _calculate_marginal_linear_model(
        self,
        log_priors: List[float],
        indicators: List[List[int]],
        models: List[ModelInfo],
    ) -> List[float]:
        log_joint_probabilities = [
            p + m.log_likelihood for (p, m) in zip(log_priors, models)
        ]
        log_marginal_likelihood = self._calculate_log_marginal_likelihood(
            log_priors=log_priors, models=models
        )
        np_coefficient = np.zeros((len(models), self.n_features_in_))
        for i, (m, indicator) in enumerate(zip(models, indicators)):
            np_coefficient[i, np.array(indicator) == 1] = m.coefficient

        coefficient = []
        for column in range(np_coefficient.shape[1]):
            coefficient_values = np_coefficient[:, column]
            result = logsumexp(
                a=log_joint_probabilities, b=coefficient_values, return_sign=True
            )
            coefficient.append(result[1] * np.exp(result[0] - log_marginal_likelihood))

        return coefficient

    def _fixed_alpha_prior(self, indicator: List[int]) -> float:
        """
        Model prior with fixed alpha:
        p(c) = prod_{i=1}^p alpha^{c_i} (1-alpha)^{1-c_i}

        Since we exclude null model, normalization constant is
        1 - (1 - alpha)^{n_features} .
        """
        n_features = len(indicator)
        n_in_use = sum(indicator)
        log_model_prior = n_in_use * np.log(self.alpha) + (
            n_features - n_in_use
        ) * np.log(1 - self.alpha)
        log_model_prior -= np.log(1 - (1 - self.alpha) ** n_features)
        return log_model_prior

    def _generate_indicator(self, n_features: int) -> List[List[int]]:
        """
        Parameters
        ----------
        n_features : int
            Number of features.

        exclude_null: bool
            Whether to exclude null model ([0, 0, .., 0]).

        Returns
        -------
        indicators : List with shape (n_combinations, n_features)
            All combinations of indicator vector.
            n_combinations is 2**n_features - 1 because null model is excluded.
        """
        offset = 1 if self.exclude_null else 0
        indicators = [list(p)[::-1] for p in product([0, 1], repeat=n_features)]
        return indicators[offset:]

    def _transform_indicator_to_model_index(self, indicator: List[int]) -> int:
        if len(indicator) != self.n_features_in_:
            raise ValueError(
                f"indicator should have be {self.n_features_in_}-length list"
            )
        offset = 1 if self.exclude_null else 0
        bin_expr = "".join(map(str, indicator))[::-1]
        return int(bin_expr, 2) - offset

    def predict(self, X: np.ndarray, mode: str, threshold: float = 0.5) -> np.ndarray:
        """Predict using the model

        Predict about new data.

        Parameters
        ----------
        X : np.ndarray with shape (n_data, n_features)
            Feature matrix for prediction.

        mode: str
            Mode of prediction.

        threshold: float (default: 0.5)
            Feature selection threshold used in mode 'select'.

        Returns
        -------
        y: np.ndarray
            Prediction values.
        """
        try:
            predict_mode = PredictMode[mode]
        except KeyError:
            raise ValueError(
                "Invalid mode: `{}` specified. Mode should be either of {}".format(
                    mode, list(PredictMode.__members__)
                )
            )

        if predict_mode == PredictMode.select:
            pred = self._predict_by_select(X=X, threshold=threshold)
        elif predict_mode == PredictMode.full:
            pred = self._predict_by_full(X=X)
        return pred

    def _predict_by_select(self, X: np.ndarray, threshold: float = 0.5) -> np.ndarray:
        indicator = self.select_variables(threshold=threshold)
        model = self.models_[
            self._transform_indicator_to_model_index(indicator=indicator)
        ]
        pred = np.dot(X[:, np.array(indicator) == 1], model.coefficient)
        return pred

    def _predict_by_full(self, X: np.ndarray) -> np.ndarray:
        pred = np.dot(X, self.coef_)
        return pred

    def select_variables(self, threshold: float = 0.5) -> List[int]:
        """
        Return indicator with posterior probability greater than or equal to threshold.
        """
        index = np.array(self.feature_posteriors_) >= threshold
        return index.astype(int).tolist()



from typing import List, Optional

import numpy as np
from scipy.special import logsumexp


def integrate_log_values_in_square(
    log_values: List[List[float]],
    x1: List[float],
    x2: List[float],
    weights: Optional[List[List[float]]] = None,
    expect_positive: bool = True,
):
    r"""
    Integrate box region.
    \int weight * exp(log_value) dx1 dx2

    log_values: 2 dimension array with shape (len(x1), len(x2))
        Log values to be integrated over x1 and x2 axes.

    x1: List[float]
        1st axis points.

    x2: List[float]
        2nd axis points.

    weights: Optional[List[List[float]]], default: None
        Weights to log_values.

    expect_positive: bool, default: True
        If set True, ValueError is raised when result is not positive.
    """
    validate_list_dimension(x1, dim=1, name="x1")
    validate_list_dimension(x2, dim=1, name="x2")

    np_log_values = np.array(log_values)
    if np_log_values.shape != (len(x1), len(x2)):
        raise ValueError(
            "Invalid shape, log_values: {}, axis: {}".format(
                np_log_values.shape, (len(x1), len(x2))
            )
        )

    if weights is None:
        np_weights = np.ones_like(np_log_values, dtype=float)
    else:
        np_weights = np.array(weights)
        if np_weights.shape != np_log_values.shape:
            raise ValueError(
                "Invalid shape, weights: {}, log_values: {}".format(
                    np_weights.shape, np_log_values.shape
                )
            )

    val = np_log_values - np.log(4)

    # For ease of implementation, extend axis points with edge values.
    extended_x1 = np.array([x1[0]] + x1 + [x1[-1]])
    extended_x2 = np.array([x2[0]] + x2 + [x2[-1]])
    area = np.zeros_like(np_log_values, dtype=float)
    for i in range(2):
        for j in range(2):
            area += np.outer(
                extended_x1[i + 1 : i + 1 + len(x1)] - extended_x1[i : i + len(x1)],
                extended_x2[j + 1 : j + 1 + len(x2)] - extended_x2[j : j + len(x2)],
            )
    val += np.log(area)

    result = logsumexp(val, b=np_weights, return_sign=True)
    if expect_positive:
        if result[1] <= 0:
            raise ValueError("Result is not positive.")
        return result[0]
    else:
        """(log of calculation, sign)"""
        return result


def integrate_log_values_in_line(
    log_values: List[float],
    x1: List[float],
    weights: Optional[List[float]] = None,
    expect_positive: bool = True,
):
    """
    Integrate along line.

    log_values: 1 dimension array with length len(x1)
        Log values to be integrated along x1

    x1: List[float]
        1st axis points.

    weights: Optional[List[float]], default: None
        Weights to log_values.

    expect_positive: bool, default: True
        If set True, ValueError is raised when result is not positive.
    """
    validate_list_dimension(x1, dim=1, name="x1")

    np_log_values = np.array(log_values)
    if np_log_values.shape != (len(x1),):
        raise ValueError(
            "Invalid shape, log_values: {}, axis: {}".format(
                np_log_values.shape, (len(x1),)
            )
        )

    if weights is None:
        np_weights = np.ones_like(np_log_values, dtype=float)
    else:
        np_weights = np.array(weights)
        if np_weights.shape != np_log_values.shape:
            raise ValueError(
                "Invalid shape, weights: {}, log_values: {}".format(
                    np_weights.shape, np_log_values.shape
                )
            )

    val = np_log_values - np.log(2)

    # For ease of implementation, extend axis points with edge values.
    extended_x1 = np.array([x1[0]] + x1 + [x1[-1]])
    area = np.zeros_like(np_log_values, dtype=float)
    for i in range(2):
        area += extended_x1[i + 1 : i + 1 + len(x1)] - extended_x1[i : i + len(x1)]
    val += np.log(area)

    result = logsumexp(val, b=np_weights, return_sign=True)
    if expect_positive:
        if result[1] <= 0:
            raise ValueError("Result is not positive.")
        return result[0]
    else:
        """(log of calculation, sign)"""
        return result


def validate_list_dimension(x, dim: int, name: str):
    if not isinstance(x, list):
        raise ValueError("{} must be list, received `{}`".format(name, type(x)))

    np_x = np.array(x)
    if len(np_x.shape) != dim:
        raise ValueError(
            "{} must be {}-dim list, received {}-dim".format(name, dim, len(np_x.shape))
        )



from typing import List, Optional, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from exhbma import ExhaustiveLinearRegression

DEFAULT_TITLESIZE = 30
DEFAULT_LABELSIZE = 24
DEFAULT_TICKLABELSIZE = 18


def feature_posterior(
    model: ExhaustiveLinearRegression,
    figsize: Optional[Tuple[float, float]] = None,
    color: str = "gray",
    title: Optional[str] = None,
    titlesize: float = DEFAULT_LABELSIZE,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    labelsize: float = DEFAULT_LABELSIZE,
    xticklabels=None,
    xrot: float = 90,
    yticklabels: List[float] = [0, 0.2, 0.4, 0.6, 0.8, 1],
    ticklabelsize: float = DEFAULT_TICKLABELSIZE,
    hlines: Optional[List[float]] = None,
):
    """
    Plot posterior probability that each feature is included in the model.

    Parameters
    ----------
    model: ExhaustiveLinearRegression
        A trained ExhaustiveLinearRegression model.
    """
    prob = model.feature_posteriors_
    index = np.arange(model.n_features_in_)

    if figsize is None:
        figsize = (len(prob), 8)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    ax.bar(index, prob, color=color)

    # Set upper limits slightly over the maximum for visibility.
    ax.set_ylim([0, 1.05])

    # Set labels
    if title is not None:
        ax.set_title(title, fontsize=titlesize)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=labelsize)
    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=labelsize)

    # Set tick labels
    if xticklabels is not None:
        assert len(xticklabels) == model.n_features_in_
        xticks = np.arange(len(xticklabels))
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, fontsize=ticklabelsize, rotation=xrot)
    if yticklabels is not None:
        yticks = yticklabels
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticklabels, fontsize=ticklabelsize)

    if hlines is not None:
        for hline in hlines:
            ax.axhline(hline, c="red", linestyle="--", linewidth=3)

    return fig, ax


def weight_diagram(
    model: ExhaustiveLinearRegression,
    n_top: int = 100,
    figsize: Optional[Tuple[float, float]] = None,
    column_downward: bool = True,
    cmap: str = "RdBu_r",
    vmax: Optional[float] = None,
    vmin: Optional[float] = None,
    title: Optional[str] = None,
    titlesize: float = DEFAULT_LABELSIZE,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    cbarlabel: Optional[str] = None,
    labelsize: float = DEFAULT_LABELSIZE,
    yticklabels: Optional[List[str]] = None,
    ticklabelsize: float = DEFAULT_TICKLABELSIZE,
):
    """
    Plot weight diagram of trained exhaustive search model.

    Parameters
    ----------
    model: ExhaustiveLinearRegression
        A trained ExhaustiveLinearRegression model.
    """

    coefs = np.zeros_like(model.indicators_, dtype=float)
    for i in range(len(model.models_)):
        coefs[i, np.array(model.indicators_[i]) == 1] = np.array(
            model.models_[i].coefficient
        )
    mcoefs = np.ma.masked_where(np.array(model.indicators_) == 0, coefs)

    if figsize is None:
        figsize = (24, model.n_features_in_)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    sorted_index = np.argsort(model.log_likelihoods_)[::-1][:n_top]
    if vmax is None and vmin is None:
        vrange = np.abs(coefs[sorted_index, :]).max()
        vmax, vmin = vrange, -vrange
    else:
        if vmax is None and vmin is not None:
            assert vmin < 0
            vmax = -vmin
        if vmax is not None and vmin is None:
            assert vmax > 0
            vmin = -vmax

    mesh = ax.pcolormesh(
        np.transpose(mcoefs[sorted_index])[:: 1 - 2 * int(column_downward)],
        cmap=cmap,
        vmax=vmax,
        vmin=vmin,
    )

    # Set labels
    if title is not None:
        ax.set_title(title, fontsize=titlesize)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=labelsize)
    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=labelsize)
    if cbarlabel is not None:
        cbar = fig.colorbar(mesh, ax=ax)
        cbar.set_label(cbarlabel, fontsize=labelsize)

    # Set tick labels
    if yticklabels is not None:
        yticks = np.arange(len(yticklabels)) + 0.5
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticklabels, fontsize=ticklabelsize)

    xticks = np.append([1], np.arange(20, n_top + 1, 20))
    xticklabels = [str(x) for x in xticks]
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels, fontsize=ticklabelsize)

    return fig, ax


def sigma_posterior(
    model: ExhaustiveLinearRegression,
    plot_range: float = 10,
    figsize: Optional[Tuple[float, float]] = None,
    cmap=None,
    title: Optional[str] = None,
    titlesize: float = DEFAULT_LABELSIZE,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    cbarlabel: Optional[str] = None,
    labelsize: float = DEFAULT_LABELSIZE,
    xticks=None,
    xticklabels=None,
    yticks=None,
    yticklabels=None,
    ticklabelsize: float = DEFAULT_TICKLABELSIZE,
):
    """
    Plot posterior probability of hyperparameter (sigma_w, sigma_epsilon).

    Parameters
    ----------
    model: ExhaustiveLinearRegression
        A trained ExhaustiveLinearRegression model.
    """
    log_prior = np.log([p.prob for p in model.sigma_noise_points]).reshape(
        -1, 1
    ) + np.log([p.prob for p in model.sigma_coef_points]).reshape(1, -1)
    log_likelihood = np.array(model.log_likelihood_over_sigma_)
    distribution = log_prior + log_likelihood - model.log_likelihood_
    threshold = distribution.max() - plot_range
    mdistribution = np.ma.masked_where(distribution < threshold, distribution)

    if figsize is None:
        xvals = np.log10([p.position for p in model.sigma_coef_points])
        xmin, xmax = min(xvals), max(xvals)
        yvals = np.log10([p.position for p in model.sigma_noise_points])
        ymin, ymax = min(yvals), max(yvals)
        figsize = ((xmax - xmin) * 3 + 2, (ymax - ymin) * 3)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    if cmap is None:
        cmap_list = [[0, "White"], [1, "#000000"]]
        cmap = LinearSegmentedColormap.from_list("mycmap", cmap_list)

    mesh = ax.pcolormesh(mdistribution, cmap=cmap)

    # Set labels
    if title is not None:
        ax.set_title(title, fontsize=titlesize)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=labelsize)
    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=labelsize)
    if cbarlabel is not None:
        cbar = fig.colorbar(mesh, ax=ax, pad=0.02)
        cbar.set_label(cbarlabel, fontsize=labelsize)

    # Set tick labels
    if xticklabels is None:
        xticks, xticklabels = _make_log_labels(
            values=[p.position for p in model.sigma_coef_points]
        )
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, fontsize=ticklabelsize)
    if yticks is None:
        yticks, yticklabels = _make_log_labels(
            values=[p.position for p in model.sigma_noise_points]
        )
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticklabels, fontsize=ticklabelsize)

    return fig, ax


def _make_log_labels(values) -> Tuple[List[float], List[str]]:
    vmin, vmax = min(values), max(values)
    ticks, ticklabels = [], []
    for i in range(int(np.log10(vmin)) - 1, int(np.log10(vmax)) + 1):
        val = 10 ** i
        if vmin <= val <= vmax:
            fmt = f"$10^{{{i}}}$" if i not in [0, 1] else str(10 ** i)
            index = np.searchsorted(values, val)
            ticks.append(index + 0.5)
            ticklabels.append(fmt)
    return ticks, ticklabels



import logging
from typing import List

import numpy as np

from exhbma.integrate import integrate_log_values_in_square
from exhbma.probabilities import RandomVariable

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class LinearRegression(object):
    """
    Model description:

    - Base model: Linear regression model
    - Observation noise: Gaussian distribution
    - Prior distribution for coefficient: Gaussian distribution
    - Intercept of the linear model is assumed to be zero.

        - Assume that target variable y is centralized.
        - Assume that all features x are centralized and normalized.
        - Marginalization over intercept is performed.

    Parameters
    ----------
    sigma_noise: float
        Standard deviation of gaussian noise.
        Model assumes that observation noise obeys Gaussian distribution
        with mean = 0, variance = sigma_noise^2.

    sigma_coef: float
        Standard deviation of gaussian noise.
        Model assumes that each coefficient value obeys Gaussian distribution
        with mean = 0, variance = sigma_coef^2 independently.

    Attributes
    ----------
    n_features_in_: int
        Number of features seen during fit.

    coef_: List[float]
        Coefficients of the regression model (mean of distribution).

    log_likelihood_: float
        Log-likelihood of the model.
        Marginalization is performed over sigma_noise, sigma_coef, indicators.
    """

    def __init__(self, sigma_noise: float, sigma_coef: float):
        self.sigma_noise = sigma_noise
        self.sigma_coef = sigma_coef
        self._preprocessing_tolerance = 1e-8

    def fit(
        self, X: np.ndarray, y: np.ndarray, skip_preprocessing_validation: bool = False
    ):
        """
        Calculate coefficient used in prediction and log likelihood for this data.

        Parameters
        ----------
        X : np.ndarray with shape (n_data, n_features)
            Feature matrix. Each row corresponds to single data.

        y : np.ndarray with shape  (n_data,)
            Target value vector.
        """
        self._validate_training_data_shape(X, y)
        if not skip_preprocessing_validation:
            self.validate_target_centralization(
                y, tolerance=self._preprocessing_tolerance
            )
            self.validate_feature_standardization(
                X, tolerance=self._preprocessing_tolerance
            )
        self.n_features_in_ = X.shape[1]

        u, s, vh = np.linalg.svd(X, full_matrices=True)
        uTy = np.dot(u.T, y)
        XTy = np.dot(X.T, y)
        vhXTy = np.dot(vh, XTy)

        self._fit_after_svd(uTy=uTy, s=s, vh=vh, vhXTy=vhXTy, y=y)

    def _fit_after_svd(
        self,
        uTy: np.ndarray,
        s: np.ndarray,
        vh: np.ndarray,
        vhXTy: np.ndarray,
        y: np.ndarray,
    ):
        """
        Calculate coefficient and log-likelihood using SVD component of X.
        """
        mu = self._calculate_coefficient(vh=vh, s=s, vhXTy=vhXTy)
        self.coef_ = mu.tolist()

        self.log_likelihood_ = self._calculate_log_likelihood(uTy=uTy, s=s, y=y)

    def _calculate_coefficient(self, s: np.ndarray, vh: np.ndarray, vhXTy: np.ndarray):
        """
        Calculate coefficient using SVD component of X.
        X = u @ np.diag(s) @ vh

        Lambda = vh.T (np.diag(s)**2 / sigma_noise**2 + 1/sigma_coef**2) vh
        mu = Lambda**(-1) X.T y / sigma_noise**2
           = vh.T (np.diag(s) ** 2 + sigma_noise**2 / sigma_coef ** 2)**(-1) vh X.T y
        """
        n_features = vh.shape[1]
        eigvals_XTX = np.zeros(n_features)
        eigvals_XTX[: len(s)] = s ** 2
        # Put sigma_noise into eigvals_lambda
        eigvals_lambda = eigvals_XTX + self.sigma_noise ** 2 / self.sigma_coef ** 2

        mu = np.dot(vh.T, vhXTy / eigvals_lambda)
        return mu

    def _calculate_log_likelihood(self, uTy: np.ndarray, s: np.ndarray, y: np.ndarray):
        """
        Calculate log likelihood using SVD component of X.
        X = u @ np.diag(s) @ vh

        log p = - N/2 log(2 pi)
                - 1/2 log det (sigma_noise**2 I + sigma_coef**2 X XT)
                - 1/2 yT (sigma_noise**2 I + sigma_coef**2 X XT)**(-1) y
                + 1/2 log(sigma_noise**2 / (N * var(y) + sigma_noise**2))

        sigma_noise**2 I + sigma_coef**2 X XT
        = u (sigma_noise**2 I + sigma_coef**2 np.diag(s)**2) u.T
        """
        n_data = len(y)
        eigvals_XXT = np.zeros(n_data)
        eigvals_XXT[: len(s)] = s ** 2
        eigvals_cov = self.sigma_noise ** 2 + self.sigma_coef ** 2 * eigvals_XXT

        const = -n_data / 2 * np.log(2 * np.pi)
        log_det = -1 / 2 * np.log(eigvals_cov).sum()
        log_exp = -1 / 2 * np.dot(uTy, uTy / eigvals_cov)
        var_y = np.var(y)
        log_intercept = (
            np.log(self.sigma_noise ** 2)
            - np.log(n_data * var_y + self.sigma_noise ** 2)
        ) / 2
        log_likelihood = const + log_det + log_exp + log_intercept
        return log_likelihood

    @staticmethod
    def _validate_training_data_shape(X, y):
        """
        Validate training data X and y.
        Check list:
        - X is 2 dimension array.
        - X and y have same number of data.
        """
        if len(X.shape) != 2:
            raise ValueError(
                "X is expect to be 2-dim array. Actual {}-dim.".format(len(X.shape))
            )
        if len(X) != len(y):
            raise ValueError(
                "Data sizes are different between X({}) and y({}).".format(
                    len(X), len(y)
                )
            )

    @staticmethod
    def validate_feature_standardization(X, tolerance: float = 1e-8):
        for i in range(X.shape[1]):
            x = X[:, i]
            x_mean = np.mean(x)
            x_var = np.var(x)
            if np.abs(x_mean) > tolerance:
                logger.error(
                    "Mean of %s-th feature(%s) is out of tolerance(%s)",
                    i,
                    x_mean,
                    tolerance,
                )
                raise ValueError(f"Feature in column-{i} is not centralized.")
            if np.abs(1 - x_var) > tolerance:
                logger.error(
                    "Variance of %s-th feature(%s) is out of tolerance(%s)",
                    i,
                    x_var,
                    tolerance,
                )
                raise ValueError(f"Feature in column-{i} is not normalized.")

    @staticmethod
    def validate_target_centralization(y, tolerance: float = 1e-8):
        y_mean = np.mean(y)
        if np.abs(y_mean) > tolerance:
            logger.error(
                "Mean of target variable(%s) is out of tolerance(%s)",
                y_mean,
                tolerance,
            )
            raise ValueError("Target variable is not centralized.")

    def predict(self, X):
        """
        Prediction using trained model.

        X : np.ndarray with shape (n_data, n_features)
            Feature matrix for prediction.
        """
        pred = np.dot(X, self.coef_)
        return pred


class ConstantRegression(object):
    """
    Model description:

    - Base model: Constant model
    - Observation noise: Gaussian distribution
    - Prior distribution for coefficient: Gaussian distribution
    - Intercept of the linear model is assumed to be zero.

        - Assume that target variable y is centralized.
        - Assume that all features x are centralized and normalized.
        - Marginalization over intercept is performed.

    Parameters
    ----------
    sigma_noise: float
        Standard deviation of gaussian noise.
        Model assumes that observation noise obeys Gaussian distribution
        with mean = 0, variance = sigma_noise^2.

    sigma_coef: float
        Standard deviation of gaussian noise.
        Model assumes that each coefficient value obeys Gaussian distribution
        with mean = 0, variance = sigma_coef^2 independently.

    Attributes
    ----------
    n_features_in_: int
        Number of features seen during fit.

    coef_: List[float]
        Coefficients of the regression model (mean of distribution).

    log_likelihood_: float
        Log-likelihood of the model.
        Marginalization is performed over sigma_noise, sigma_coef, indicators.
    """

    def __init__(self, sigma_noise: float, sigma_coef: float):
        self.sigma_noise = sigma_noise
        self.sigma_coef = sigma_coef
        self._preprocessing_tolerance = 1e-8

    def fit(
        self, X: np.ndarray, y: np.ndarray, skip_preprocessing_validation: bool = False
    ):
        """
        Calculate log likelihood for this data.

        Parameters
        ----------
        X : np.ndarray with shape (n_data, n_features)
            Feature matrix. Each row corresponds to single data.

        y : np.ndarray with shape  (n_data,)
            Target value vector.
        """
        if not skip_preprocessing_validation:
            LinearRegression.validate_target_centralization(
                y, tolerance=self._preprocessing_tolerance
            )
            LinearRegression.validate_feature_standardization(
                X, tolerance=self._preprocessing_tolerance
            )
        self.n_features_in_ = 0

        self.coef_: List[float] = []
        self.log_likelihood_ = self._calculate_log_likelihood(y=y)

    def _calculate_log_likelihood(self, y: np.ndarray):
        """
        Calculate log Likelihood for constant model.
        log p = - N/2 log(2 pi * sigma_noise**2)
                - 1/2 yTy**2 / sigma_noise**2
                + 1/2 log(sigma_noise**2 / (N var(y) + sigma_noise**2))
        """
        n_data = len(y)

        const = -n_data / 2 * (np.log(2 * np.pi) + np.log(self.sigma_noise ** 2))
        log_exp = -1 / 2 * np.dot(y, y) / self.sigma_noise ** 2
        var_y = np.var(y)
        log_intercept = (
            np.log(self.sigma_noise ** 2)
            - np.log(n_data * var_y + self.sigma_noise ** 2)
        ) / 2
        log_likelihood = const + log_exp + log_intercept
        return log_likelihood

    def predict(self, X):
        return 0


class MarginalLinearRegression(object):
    r"""
    Model description:

    - Base model: Linear regression model
    - Observation noise: Gaussian distribution
    - Prior distribution for coefficient: Gaussian distribution
    - Intercept of the linear model is assumed to be zero.
    - Marginalization over sigma_noise and sigma_coefficient are performed.

        - Assume that target variable y is centralized.
        - Assume that all features x are centralized and normalized.
        - Marginalization over intercept is performed.

    Parameters
    ----------
    sigma_noise_points: List[RandomVariable]
        Data points to marginalize over sigma_noise parameter.

    sigma_coef_points: List[RandomVariable]
        Data points to marginalize over sigma_coef parameter.


    Attributes
    ----------
    n_features_in_: int
        Number of features seen during fit.

    coef_: List[float]
        Coefficients of the regression model (mean of distribution).

    log_likelihood_: float
        Log-likelihood of the model.
        Marginalization is performed over sigma_noise, sigma_coef, indicators.

    log_likelihood_over_sigma_: List[List[float]]
        Log-likelihood over :math:`\sigma_{noise}` and :math:`\sigma_{coef}`,
        :math:`p(y| \sigma_{noise}, \sigma_{coef}, X)`.
        Prior distributions for both sigma are not included.
    """

    def __init__(
        self,
        sigma_noise_points: List[RandomVariable],
        sigma_coef_points: List[RandomVariable],
    ):
        self.sigma_noise_points = sigma_noise_points
        self.sigma_coef_points = sigma_coef_points
        self._preprocessing_tolerance = 1e-8

    def fit(
        self, X: np.ndarray, y: np.ndarray, skip_preprocessing_validation: bool = False
    ):
        """
        Calculate coefficient used in prediction and log likelihood for this data.

        Parameters
        ----------
        X : np.ndarray with shape (n_data, n_features)
            Feature matrix. Each row corresponds to single data.

        y : np.ndarray with shape  (n_data,)
            Target value vector.
        """
        LinearRegression._validate_training_data_shape(X, y)
        if not skip_preprocessing_validation:
            LinearRegression.validate_target_centralization(
                y, tolerance=self._preprocessing_tolerance
            )
            LinearRegression.validate_feature_standardization(
                X, tolerance=self._preprocessing_tolerance
            )
        self.n_features_in_ = X.shape[1]

        # Fit models
        fit_models: List[List[LinearRegression]] = self._fit_models_over_sigma(X=X, y=y)

        # Calculate log likelihood
        log_likelihood_over_sigma = np.array(
            [
                [model.log_likelihood_ for model in models_along_coef]
                for models_along_coef in fit_models
            ]
        )
        log_prior = np.log([p.prob for p in self.sigma_noise_points]).reshape(
            -1, 1
        ) + np.log([p.prob for p in self.sigma_coef_points]).reshape(1, -1)
        log_joint_probabilities = log_likelihood_over_sigma + log_prior

        log_likelihood = self._integrate_log_likelihood_over_sigma(
            log_joint_probabilities=log_joint_probabilities
        )

        # Integrate coefficient by log_prob_values
        coefficient = []
        for i in range(X.shape[1]):
            coefficient.append(
                self._integrate_coefficient_over_sigma(
                    index=i,
                    fit_models=fit_models,
                    log_joint_probabilities=log_joint_probabilities,
                    log_likelihood=log_likelihood,
                )
            )

        self.log_likelihood_ = log_likelihood
        self.log_likelihood_over_sigma_ = log_likelihood_over_sigma.tolist()
        self.coef_ = coefficient

    def _fit_models_over_sigma(self, X, y) -> List[List[LinearRegression]]:
        u, s, vh = np.linalg.svd(X, full_matrices=True)
        uTy = np.dot(u.T, y)
        XTy = np.dot(X.T, y)
        vhXTy = np.dot(vh, XTy)

        fit_models: List[List[LinearRegression]] = []
        for sigma_noise in self.sigma_noise_points:
            models_along_coef = []
            for sigma_coef in self.sigma_coef_points:
                reg = LinearRegression(
                    sigma_noise=sigma_noise.position, sigma_coef=sigma_coef.position
                )
                reg._fit_after_svd(uTy=uTy, s=s, vh=vh, vhXTy=vhXTy, y=y)
                models_along_coef.append(reg)
            fit_models.append(models_along_coef)
        return fit_models

    def _integrate_log_likelihood_over_sigma(
        self,
        log_joint_probabilities,
    ) -> float:
        # Marginalize over sigma_noise and sigma_coef
        log_likelihood = integrate_log_values_in_square(
            log_values=log_joint_probabilities.tolist(),
            x1=[p.position for p in self.sigma_noise_points],
            x2=[p.position for p in self.sigma_coef_points],
        )
        return log_likelihood

    def _integrate_coefficient_over_sigma(
        self,
        index: int,
        fit_models: List[List[LinearRegression]],
        log_joint_probabilities,
        log_likelihood: float,
    ) -> float:
        coefficient_weights = [
            [model.coef_[index] for model in models_along_coef]
            for models_along_coef in fit_models
        ]
        result = integrate_log_values_in_square(
            log_values=(log_joint_probabilities - log_likelihood).tolist(),
            x1=[p.position for p in self.sigma_noise_points],
            x2=[p.position for p in self.sigma_coef_points],
            weights=coefficient_weights,
            expect_positive=False,
        )
        return result[1] * np.exp(result[0])

    def predict(self, X):
        """
        Prediction using trained model.

        X : np.ndarray with shape (n_data, n_features)
            Feature matrix for prediction.
        """
        pred = np.dot(X, self.coef_)
        return pred


class MarginalNullRegression(object):
    r"""
    Model description:

    - Base model: Constant model
    - Observation noise: Gaussian distribution
    - Intercept of the linear model is assumed to be zero.
    - Marginalization over sigma_noise and sigma_coefficient are performed.

        - Assume that target variable y is centralized.
        - Assume that all features x are centralized and normalized.
        - Marginalization over intercept is performed.

    Parameters
    ----------
    sigma_noise_points: List[RandomVariable]
        Data points to marginalize over sigma_noise parameter.

    sigma_coef_points: List[RandomVariable]
        Data points to marginalize over sigma_coef parameter.


    Attributes
    ----------
    n_features_in_: int
        Number of features seen during fit.

    coef_: List[float]
        Coefficients of the regression model (mean of distribution).

    log_likelihood_: float
        Log-likelihood of the model.
        Marginalization is performed over sigma_noise, sigma_coef, indicators.

    log_likelihood_over_sigma_: List[List[float]]
        Log-likelihood over :math:`\sigma_{noise}` and :math:`\sigma_{coef}`,
        :math:`p(y| \sigma_{noise}, \sigma_{coef}, X)`.
        Prior distributions for both sigma are not included.
    """

    def __init__(
        self,
        sigma_noise_points: List[RandomVariable],
        sigma_coef_points: List[RandomVariable],
    ):
        self.sigma_noise_points = sigma_noise_points
        self.sigma_coef_points = sigma_coef_points
        self._preprocessing_tolerance = 1e-8

    def fit(
        self, X: np.ndarray, y: np.ndarray, skip_preprocessing_validation: bool = False
    ):
        """
        Calculate log likelihood for this data.

        Parameters
        ----------
        X : np.ndarray with shape (n_data, 0)
            Null feature matrix.

        y : np.ndarray with shape  (n_data,)
            Target value vector.
        """
        LinearRegression._validate_training_data_shape(X, y)
        if not skip_preprocessing_validation:
            LinearRegression.validate_target_centralization(
                y, tolerance=self._preprocessing_tolerance
            )
            LinearRegression.validate_feature_standardization(
                X, tolerance=self._preprocessing_tolerance
            )
        self.n_features_in_ = X.shape[1]

        # Fit models
        fit_models: List[List[ConstantRegression]] = self._fit_models_over_sigma(
            X=X, y=y
        )

        # Calculate log likelihood
        log_likelihood_over_sigma = np.array(
            [
                [model.log_likelihood_ for model in models_along_coef]
                for models_along_coef in fit_models
            ]
        )
        log_prior = np.log([p.prob for p in self.sigma_noise_points]).reshape(
            -1, 1
        ) + np.log([p.prob for p in self.sigma_coef_points]).reshape(1, -1)
        log_joint_probabilities = log_likelihood_over_sigma + log_prior

        log_likelihood = self._integrate_log_likelihood_over_sigma(
            log_joint_probabilities=log_joint_probabilities
        )

        self.log_likelihood_ = log_likelihood
        self.log_likelihood_over_sigma_ = log_likelihood_over_sigma.tolist()
        self.coef_: List[float] = []

    def _fit_models_over_sigma(self, X, y) -> List[List[ConstantRegression]]:
        fit_models: List[List[ConstantRegression]] = []
        for sigma_noise in self.sigma_noise_points:
            models_along_coef = []
            for sigma_coef in self.sigma_coef_points:
                reg = ConstantRegression(
                    sigma_noise=sigma_noise.position, sigma_coef=sigma_coef.position
                )
                reg.fit(X=X, y=y, skip_preprocessing_validation=True)
                models_along_coef.append(reg)
            fit_models.append(models_along_coef)
        return fit_models

    def _integrate_log_likelihood_over_sigma(
        self,
        log_joint_probabilities,
    ) -> float:
        # Marginalize over sigma_noise and sigma_coef
        log_likelihood = integrate_log_values_in_square(
            log_values=log_joint_probabilities.tolist(),
            x1=[p.position for p in self.sigma_noise_points],
            x2=[p.position for p in self.sigma_coef_points],
        )
        return log_likelihood

    def predict(self, X):
        return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exhbma']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.1,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'exhbma',
    'version': '0.1.8',
    'description': 'Exhaustive Search with Bayesian Model Averaging',
    'long_description': '# Exhaustive Search with Bayesian Model Averaging (ExhBMA)\n\n# Installation\n```\npip install exhbma\n```\n\n# Documentation\nUser documentation is available [here](https://exhbma.readthedocs.io).\nYou can try sample notebooks in [tutorials](/tutorials) directory.\n\n# Reference paper\nIn preparing.\n',
    'author': 'Koki Obinata',
    'author_email': 'koki.obi.321@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/okada-lab/exhbma',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
